// ************************************
// Brain Stimulator
// Version: 76 (beta)
//
// Changelog:
//
// January 30, 2014 -- Brian Neltner
// Created Serial Receive Routine
// Added DAC, DDS, and WSS modes.
//
// December 7, 2017 -- Brian Neltner
// Removed all intermediate changelog comments from header to reduce clutter. See previous versions for comments.
//
// March 5, 2018 -- Brian Neltner
// Adding stimulation program uploading capability.
//
// ************************************

// The title screen text.
#define title "Trans Cranial Stimulator 24 Channel v79"

#include <SPI.h>
#include <vector>
#include <LiquidCrystal.h>
#include "StimPinConfig.h"
#include "stimulatorInterface.h"

// Set default DDS frequencies.
#define DDS0freq 10
#define DDS1freq 10

// Set up a vector table for stimulation program storage.
unsigned long stimulationstepstarttime;
unsigned long stimulationstependtime;
unsigned long stimulationstepduration;

// Set up a flag to indicate whether we're currently running a program.
boolean stimulationrunning = false;

// This will always point to the currently running row of a stimulation table.
unsigned int stimulationrownum = 0;

struct stimulationrow {
  float preidletime; // The time since stimulation start that the row should be started.
  float rampuptime;
  float duration;
  float rampdowntime;
  float postidletime;
  float freq0;
  float amp0;
  float offset0;
  float freq1;
  float amp1;
  float offset1;
  unsigned int ch1load;
  unsigned int ch2load;
  unsigned int ch3load;
  unsigned int ch4load;
  unsigned int ch1pin;
  unsigned int ch2pin;
  unsigned int ch3pin;
  unsigned int ch4pin;
  unsigned int trigger;
  unsigned int chconfig;
};

std::vector<stimulationrow> stimulationtable;

// And this is our state machine for running a stimulation.
enum stimulationtablestate {
  SETUP,
  PREIDLE,
  RAMPUP,
  STIMRUN,
  RAMPDOWN,
  POSTIDLE
} stimulationtablestate;

float currentDDS0freq, currentDDS1freq;

// Default ChConfig
unsigned char ChConfig[8] = {0b0010, 0b0010, 0b0010, 0b0010, 0, 1, 2, 3};
// Sets all loads to 10k resistor, and pins to 1, 2, 3, and 4.
unsigned char ChConfigSave[8];
float Wfm0Current, Wfm1Current, Wfm0Save, Wfm1Save;
enum stimulationstate stimulationstate;
boolean streamingFlag = 0;
boolean EEGLockingFlag = false;

StimulatorInterface stimInterface;

volatile float EEGFreq, EEGAmp, EEGPhase;
volatile boolean EEGPhaseUpdated = false;
volatile unsigned long EEGPhaseTime;

void onDataFunction(float freq_, float amp_, float phase_){
  EEGPhaseTime = micros();
  EEGFreq = freq_;
  EEGAmp = amp_;
  EEGPhase = phase_;
  EEGPhaseUpdated = true;
}

String LoadName[17] = {"OpenCkt","Electrd","Res 10k","Ele+Res","Cap10nF","Ele+Cap","Res+Cap","E+R+Cap","Ground ","Ele+Gnd","Cap+Gnd","E+R+Gnd","E+C+Gnd","Res+Gnd","E+R+Gnd","R+C+Gnd","E+R+C+G"}; 

uint8_t analog[31998]; // The ADC buffer.
unsigned int analogpointer = 12; // Points to current buffer location.

unsigned long lastmicros;

void setup() {
  // Delay for 1 second on boot to allow ICs to become stable.
  delay(100);
  stimulationstate = RUN;

  // Set up Arduino pins as outputs and inputs respectively.
  initPins();
  initAnalog();

  // Set default loads and pins for multiplexer.
  SendCh1Load(2); // for 10kOhm
  SendCh2Load(2); // for 10kOhm
  SendCh3Load(2); // for 10kOhm
  SendCh4Load(2); // for 10kOhm
  SetCh1Electrode(0);
  SetCh2Electrode(1);
  SetCh3Electrode(2);
  SetCh4Electrode(3);

  // Set default Wfm0 and Wfm1 source
  SetWSS(0, "SIN0"); // Set Wfm0 output to DDS0.
  SetWSS(1, "SIN1"); // Set Wfm1 output to DDS1.

  initSerial();

  // set up the LCD's number of columns and rows:
  lcd.begin(40, 2);
  lcd2.begin(40, 2);
  lcd.clear();
  lcd2.clear();
  lcd2.print("");
  lcd.setCursor(0, 0);
  lcd.print(title);
  lcd.display();
  lcd2.display();
  float voltage22plus, voltage22minus, voltagebattery, voltage165;
  voltage22plus = (float)analogRead(8) * 3.3 / 0xFFFF * 21.0;
  voltage22minus = (float)analogRead(9) * 3.3 / 0xFFFF * 21.0 - 66.0;
  voltagebattery = (float)analogRead(10) * 3.3 / 0xFFFF * 10.0;
  voltage165 = (float)analogRead(11) * 3.3 / 0xFFFF;
  lcd.setCursor(0, 1);
  lcd.print(("Battery:  " + String(voltagebattery)));
  lcd.setCursor(20, 1);
  lcd.print(("Vref:  " + String(voltage165)));
  lcd2.setCursor(0, 0);
  lcd2.print(("     -V: " + String(voltage22minus)));
  lcd2.setCursor(20, 0);
  lcd2.print(("  +V: " + String(voltage22plus)));
  delay(3000);
  lcd.setCursor(0, 0);
  lcd.print("                                        ");
  lcd.setCursor(0, 1);
  lcd.print("                                        ");
  lcd2.setCursor(0, 0);
  lcd2.print("                                        ");
  lcd2.setCursor(0, 1);
  lcd2.print("                                        ");

  // Initialize DAC outputs.
  SetDAC("A", 1);
  SetDAC("B", -1);
  SetDAC("C", 0);
  SetDAC("D", 1.65);
  SetDAC("E", -1.65); // suitable for DDS use as default signal source
  SetDAC("F", -1.65);
  SetDAC("G", 0);
  SetDAC("H", 0);

  // Set DDS0 and DDS1 Phases to zero using independent registers.
  SetDDS0phase(0);
  SetDDS1phase(0);

  // Initialize DDS0 and DDS1
  SetDDS0freq(DDS0freq); // Set DDS0 to default.
  SetDDS1freq(DDS1freq); // Set DDS1 to default.

  

  // Set default Wfm0 and Wfm1 current
  lcd.setCursor(0, 1);
  lcd.print("Wfm 0 mA: ");
  lcd.print(0);
  SetMDAC0(0);
  lcd.setCursor(20, 1);
  lcd.print("Wfm 1 mA: ");
  lcd.print(0);
  SetMDAC1(0); // Set Wfm1 output to 0 mA.

  updateLCDLoads();

  // This sets up packet framing. 0xFFFFFFFFFFFF is impossible to receive because the ADC values are only 12-bit.
  // This will indicate the start of a data packet to LabView.
  for (int i=0; i<6; i++) {
    analog[i] = 0xFF;
  }
}

void updateLCDLoads(void) {
  lcd2.setCursor(0, 0);
  lcd2.print("CH1: ");

  int ElectrodeNumber[4];
  for (byte i = 0; i < 4; i++) {
    ElectrodeNumber[i] = ChConfig[i+4];
    if (ElectrodeNumber[i] < 24) ElectrodeNumber[i] = ElectrodeNumber[i] - 7;
    else if (ElectrodeNumber[i] >= 32 && ElectrodeNumber[i] < 40) ElectrodeNumber[i] = ElectrodeNumber[i] - 15;
  }
  
  if (ElectrodeNumber[0]  < 10) lcd2.print("0");
  lcd2.print(ElectrodeNumber[0]);
  lcd2.print("   CH2: ");
  if (ElectrodeNumber[1]  < 10) lcd2.print("0");
  lcd2.print(ElectrodeNumber[1]);
  lcd2.print("   CH3: ");
  if (ElectrodeNumber[2]  < 10) lcd2.print("0");
  lcd2.print(ElectrodeNumber[2]);
  lcd2.print("   CH4: ");
  if (ElectrodeNumber[3]  < 10) lcd2.print("0");
  lcd2.print(ElectrodeNumber[3]);
  lcd2.setCursor(0, 1);
  // lcd2.print("");
  lcd2.print(LoadName[ChConfig[0]]);
  lcd2.print("   ");
  lcd2.print(LoadName[ChConfig[1]]);
  lcd2.print("   ");
  lcd2.print(LoadName[ChConfig[2]]);
  lcd2.print("   ");
  lcd2.print(LoadName[ChConfig[3]]);
}

boolean sendFlag = 0;

void loop() {
  if (SerialUSB.available()) {
    evaluatecommand(SerialUSB.readStringUntil(0x0D));
  }
  
  if (EEGLockingFlag) {
    // Update freq, amp, and phase global variables using the callback method "OPTION 2".
    // I added printing the data to the USB Serial port inside the callback.
    stimInterface.update();
  }

  unsigned long currentTime = micros();

  // Code to handle a stimulation table.
  if (stimulationrunning) { // This is the flag indicating that stimulation is active.
    switch (stimulationtablestate) {
      case SETUP:
        SetDDS0freq(stimulationtable[stimulationrownum].freq0); // Set DDS0 frequency.
        SetDDS1freq(stimulationtable[stimulationrownum].freq1); // Set DDS1 frequency.
        SetMDAC0(0); // Set current 0 output to zero for start.
        SetMDAC1(0); // Set current 1 output to zero for start.
        SetWSS(0, "SIN0"); // Set waveform 0 to sin0.
        SetWSS(1, "SIN1"); // Set waveform 1 to sin1.
        stimulationstependtime = stimulationstepstarttime + stimulationtable[stimulationrownum].preidletime * 1000000; // Add the preidletime in milliseconds to the current time.
        stimulationstepduration = stimulationstependtime - stimulationstepstarttime;
        stimulationtablestate = PREIDLE; // And with the setup done we move to the PREIDLE stage.
        break;
      case PREIDLE:
        if (currentTime >= stimulationstependtime) { // If the correct number of milliseconds have passed...
          SetChElectrode(1, stimulationtable[stimulationrownum].ch1pin); // Configure the electrode output pins.
          SetChElectrode(2, stimulationtable[stimulationrownum].ch2pin);
          SetChElectrode(3, stimulationtable[stimulationrownum].ch3pin);
          SetChElectrode(4, stimulationtable[stimulationrownum].ch4pin);
          SetChLoad(1, stimulationtable[stimulationrownum].ch1load); // Configure the loads.
          SetChLoad(2, stimulationtable[stimulationrownum].ch2load);
          SetChLoad(3, stimulationtable[stimulationrownum].ch3load);
          SetChLoad(4, stimulationtable[stimulationrownum].ch4load);
          updateLCDLoads();
          stimulationstepstarttime = stimulationstependtime;
          stimulationstependtime = stimulationstepstarttime + stimulationtable[stimulationrownum].rampuptime * 1000000; // Add the rampup time in milliseconds to the current time.
          stimulationstepduration = stimulationstependtime - stimulationstepstarttime;
          stimulationtablestate = RAMPUP; // And with the preidle step done we move to the RAMPUP stage.
          digitalWrite(triggerPin, stimulationtable[stimulationrownum].trigger); // Send the trigger output for this row.
        }
        break;
      case RAMPUP:
        if (currentTime >= stimulationstependtime) { // Move on if the ramp is over.
          SetMDAC0(stimulationtable[stimulationrownum].amp0); // Manually set the current to the target stimulation level.
          SetMDAC1(stimulationtable[stimulationrownum].amp1);
          stimulationstepstarttime = stimulationstependtime;
          stimulationstependtime = stimulationstepstarttime + stimulationtable[stimulationrownum].duration * 1000000; // Add the duration in milliseconds to the current time.
          stimulationstepduration = stimulationstependtime - stimulationstepstarttime;
          stimulationtablestate = STIMRUN; // And with the ramp up done we move to the STIMRUN stage.
        }
        else { // Otherwise, set to ramp output up, non-blocking.
          float scaler = (1 / (float)stimulationstepduration) * (float)(currentTime - stimulationstepstarttime);
          SetMDAC0(scaler * stimulationtable[stimulationrownum].amp0);
          SetMDAC1(scaler * stimulationtable[stimulationrownum].amp1);
        }
        break;
      case STIMRUN:
        if (currentTime >= stimulationstependtime) { // Move on if the stimulation pulse is over.
          stimulationstepstarttime = stimulationstependtime;
          stimulationstependtime = stimulationstepstarttime + stimulationtable[stimulationrownum].rampdowntime * 1000000; // Add the duration in milliseconds to the current time.
          stimulationstepduration = stimulationstependtime - stimulationstepstarttime;
          stimulationtablestate = RAMPDOWN; // And with the stimulation pulse done we move to the RAMPDOWN stage.
        }
        break;
      case RAMPDOWN:
        if (currentTime >= stimulationstependtime) { // Move on if the ramp down is over.
          SetMDAC0(0); // Set both current outputs to zero, otherwise leave the configuration as-is.
          SetMDAC1(0);
          if (stimulationtable[stimulationrownum].chconfig == 1) {
            SetChLoad(1, 0b0010); // Connect Channel Loads to the 10k resistor, disconnect from electrodes.
            SetChLoad(2, 0b0010);
            SetChLoad(3, 0b0010);
            SetChLoad(4, 0b0010);
            updateLCDLoads();
          }
          stimulationstepstarttime = stimulationstependtime;
          stimulationstependtime = stimulationstepstarttime + stimulationtable[stimulationrownum].postidletime * 1000000; // Add the duration in milliseconds to the current time.
          stimulationstepduration = stimulationstependtime - stimulationstepstarttime;
          digitalWrite(triggerPin, LOW); // Reset the trigger output for this row.
          stimulationtablestate = POSTIDLE; // And with the ramp down done we move to the POSTIDLE stage.
        }
        else { // Otherwise, set to ramp output down, non-blocking.
          float scaler = (1 / (float)stimulationstepduration) * (float)(currentTime - stimulationstepstarttime);
          SetMDAC0((1-scaler) * stimulationtable[stimulationrownum].amp0);
          SetMDAC1((1-scaler) * stimulationtable[stimulationrownum].amp1);
        }
        break;
      case POSTIDLE:
        if (currentTime >= stimulationstependtime) { // If the postidle time is done, increment the row number and go back to SETUP.
          stimulationrownum++; // Increment the row number.
          if (stimulationrownum >= stimulationtable.size()) stimulationrunning = false; // If the table is all done, turn off stimulation.
          else {
            stimulationtablestate = SETUP; // Otherwise, move to the next line's SETUP state. 
            stimulationstepstarttime = stimulationstependtime;
          }
        }
        break;
      default:
        stimulationrunning = false;
        SetMDAC0(0); // Turn off both output DACs.
        SetMDAC1(0);
        SetChLoad(1, 0b0010); // Connect Channel Loads to the 10k resistor, disconnect from electrodes.
        SetChLoad(2, 0b0010);
        SetChLoad(3, 0b0010);
        SetChLoad(4, 0b0010);
        updateLCDLoads();
        SerialUSB.println("ERROR CODE 329157: Stimulation Table Execution has a firmware bug assigning invalid state. Stimulation aborted and set to safe mode.");
        break;
    }
  }
  
  if ((streamingFlag == 1) and (currentTime > (lastmicros + 10))) {
    if (analogpointer < 31998-12) { // If the buffer has room for another data point, take it.
      unsigned int analogvalue0, analogvalue1, analogvalue4, analogvalue5;
      analogvalue0 = analogRead(0);
      analogvalue1 = analogRead(1);
      analogvalue4 = analogRead(4);
      analogvalue5 = analogRead(5);
      analog[analogpointer] = (currentTime >> 24) & 0xFF;
      analog[analogpointer + 1] = (currentTime >> 16) & 0xFF;
      analog[analogpointer + 2] = (currentTime >> 8) & 0xFF;
      analog[analogpointer + 3] = currentTime & 0xFF;
      analog[analogpointer + 4] = (analogvalue0 >> 8) & 0xFF;
      analog[analogpointer + 5] = analogvalue0 & 0xFF;
      analog[analogpointer + 6] = (analogvalue1 >> 8) & 0xFF;
      analog[analogpointer + 7] = analogvalue1 & 0xFF;
      analog[analogpointer + 8] = (analogvalue4 >> 8) & 0xFF;
      analog[analogpointer + 9] = analogvalue4 & 0xFF;
      analog[analogpointer + 10] = (analogvalue5 >> 8) & 0xFF;
      analog[analogpointer + 11] = analogvalue5 & 0xFF;
      analogpointer += 12; // And increment the pointer.
      lastmicros = currentTime;
    }
    else {
      sendFlag = 1;
    }
  }
  if (sendFlag) { // If the buffer is full, send it and reset the pointer to 6.
    SerialUSB.write(analog, 31998);
    analogpointer = 6; // Note that the first 6 bytes are 0xFF to help with framing the packet in LabView.
    sendFlag = 0;
  }
}


